/*
 * Copyright (c) 2012 AgroSense.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    AgroSense - initial API and implementation and/or initial documentation
 */
package nl.cloudfarming.client.model;

import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import org.openide.util.Lookup;
import org.openide.util.lookup.ServiceProvider;

/**
 * Centralized cached for all URI based entities.
 * 
 * @author johan
 */
public abstract class EntityCache {
    private static EntityCache instance;
    
    public static synchronized EntityCache getInstance() {
        if (instance == null) {
            instance = Lookup.getDefault().lookup(EntityCache.class);
            if (instance == null) {
                instance = new DefaultImpl();
            }
        }
        return instance;
    }
    
    public abstract void add(Entity e);
    
    // convenience method:
    public void remove(Entity e) {
        remove(e.getURI());
    }
    
    public abstract void remove(AgroURI uri);
    
    public abstract Entity get(AgroURI uri);

    // index/partition by farm/project
    public abstract Collection<AgroURI> get(FarmID farmID);
    
    public abstract void remove(FarmID farmID);
    
    public abstract void clear();
    
    
    @ServiceProvider(service=EntityCache.class)
    public static class DefaultImpl extends EntityCache {
        private final Map<AgroURI, Entity> cache = new ConcurrentHashMap<>();
        private Map<FarmID, Set<AgroURI>> idxByFarm = new ConcurrentHashMap<>();

        @Override
        public synchronized void add(Entity e) {
            // do not replace cached instances:
            if (!cache.containsKey(e.getURI())) {
                cache.put(e.getURI(), e);

                FarmID farmID = new FarmID(e.getURI());
                Set<AgroURI> uris = idxByFarm.get(farmID);
                if (uris == null) {
                    uris = new HashSet<>();
                    idxByFarm.put(farmID, uris);
                }
                uris.add(e.getURI());
            }
        }

        @Override
        public synchronized void remove(AgroURI uri) {
            cache.remove(uri);
            
            FarmID farmID = new FarmID(uri);
            Set<AgroURI> uris = idxByFarm.get(farmID);
            if (uris != null) {
                uris.remove(uri);
//                if (uris.isEmpty()) {
//                    idxByFarm.remove(farmID);
//                }
            }
        }

        @Override
        public Entity get(AgroURI uri) {
            return cache.get(uri);
        }

        @Override
        public Collection<AgroURI> get(FarmID farmID) {
            Set<AgroURI> set = idxByFarm.get(farmID);
            return set == null ? null : Collections.unmodifiableSet(set);
        }

        @Override
        public synchronized void remove(FarmID farmID) {
            Collection<AgroURI> uris = get(farmID);
            if (uris != null) {
                for (AgroURI uri : uris) {
                    cache.remove(uri);
                }
                idxByFarm.remove(farmID);
            }
        }

        @Override
        public synchronized void clear() {
            cache.clear();
            idxByFarm.clear();
        }
    
    }
}
